/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package library;

import java.util.NoSuchElementException;
import java.util.Queue;
import java.util.concurrent.ConcurrentLinkedQueue;

/**
 *
 * @author Ben
 */
public class ClockSynchronisation
{

    
    private static final int threadCount = 10;
    static Queue<ExecutionContext> currentUpdateTaskList = new ConcurrentLinkedQueue<>();
    static Queue<ExecutionContext> nextUpdateTaskList = new ConcurrentLinkedQueue<>();
    /**
     * Count of currently running threads. Should look at Thread.join() if you are ok with creating and destroying threads.
     * Has an initial value to get all the threads synchronised to the clock on init. It's entirely possible for the master thread 
     * to start the first clock cycle before the slave threads have been parked.
     */
    // must be guarded by park_master
    private static int runningThreads = 0;
    private static final Object park_master = new Object();
    
    public static int READ_BUFFER = 0;
    public static int WRITE_BUFFER = 1;
    
    
    
    public static void flip()
    {
        {
            int temp = WRITE_BUFFER;
            WRITE_BUFFER = READ_BUFFER;
            READ_BUFFER = temp;
        }
        {
            Queue<ExecutionContext> temp = currentUpdateTaskList;
            currentUpdateTaskList = nextUpdateTaskList;
            nextUpdateTaskList = temp;
        }
    }
    
    public static void run()
    {
        synchronized (park_master)
        {
            if(0 != runningThreads)
                throw new RuntimeException("There were running threads");
            runningThreads = threadCount;
        // -1 on thread count because of the "master thread"
        }
        for(int count = 0;  count < threadCount -1; count++)
        {
            new Thread(new RunProcess()).start();
//            System.out.println("Thread created");
        }
        
        while(execute_Context());
        synchronized(park_master)
        {
            runningThreads -= 1;
            if(runningThreads != 0) // there are still slaves running, wait for them to wake this.
                sleepOn(park_master);
        }
    
    }
    
    
    
    private static class RunProcess
    implements Runnable
    {
        
        @Override
        public void run()
        {
            while(execute_Context());
            synchronized (park_master)
            {
                runningThreads -= 1;
//                System.out.println("Ending: " + runningThreads);
                if(runningThreads == 0) // If it was the last thread running, wake the master thread & die.
                    wakeAll(park_master);
            }
//            System.out.println("threadEnd");
        }
    }
    
    
    
    // Gets a context from the queue and executes it. Returns if the queue was empty.
    private static boolean execute_Context()
    {
        try{
            ExecutionContext next = currentUpdateTaskList.remove();
            next.execute();
        }catch(NoSuchElementException wasEmpty)
        {
            return false;
        }
        return true;
    }
    
    
    
    private static void sleepOn(Object lock)
    {
        synchronized(lock)
        {
            try
            {
                lock.wait(); 
            }
            catch(InterruptedException ex)
            {
                // We are expecting this exception, so ignore.
            }
        }
    }
    
    
    private static void wakeAll(Object lock)
    {
        synchronized (lock)
        {
            lock.notifyAll();
        }
    }
    
    
    
    
}
